// DiscoveryServer.java

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.StringTokenizer;

public class DiscoveryServer {

  private static final int PORT = 4445;

  // porta al di fuori del range 1-1024 !
  // dichiarata come statica perch caratterizza il server

  public static void main(String[] args) {

    int validArgs = 0;

    // al massimo sono args.length/2 socket address, se tutti sono corretti
    InetAddress[] validMCastAddresses = new InetAddress[args.length / 2];    
    
    // al massimo sono args.length/2 file, se tutti sono corretti
    File[] validFiles = new File[args.length / 2];

    // al massimo sono args.length/2 file, se tutti sono corretti
    int[] validPorts = new int[args.length / 2];

    // al massimo sono i server disponibili sono pari al numero di coppie di
    // parametri di ingresso valide
    boolean[] correctlyActivated;

    /* Controllo argomenti */
    // Se non ho un numero pari di argomenti, oppure non viene registrato nessun server posso uscire
    if (args.length % 2 != 0 || args.length == 0 ) {
      System.out
          .println("Usage: java DiscoveryServer file1 MCAddr1 ... fileN MCAddrN");
      System.exit(1);
    }
    File fileCorr;
    InetAddress groupCorr;

    // per ogni coppia verifico che il primo argomento sia un file esistente
    // e il secondo un indirizzo multicast
    for (int i = 1; i < args.length; i = i + 2) {
      if (!((fileCorr = new File(args[i - 1])).exists())) {
        System.out
            .println("One of the file passed as args, does not exist, the following:\n"
                + args[i - 1]);
        continue;
      }
      try {
        groupCorr = InetAddress.getByName(args[i]);
        if (!groupCorr.isMulticastAddress()) {
          System.out
              .println("The following address in not a multicast address: "
                  + args[i]);
          continue;
        } else { //se arrivo qui: (a) il file esiste (b) l'indirizzo MCast e'
          // corretto
          // aggiorno le strutture dati
          validFiles[validArgs] = fileCorr;
          // setto solo l'IP, la porta non la conosco ancora e la setto dopo...
          validMCastAddresses[validArgs] = groupCorr;
          validArgs++;
          System.out.println("Aggiunto il seguente server multicast: ");
          System.out.println("File: " + fileCorr + ",\t IP multicast: "
              + groupCorr);
        }
      }
      catch (UnknownHostException e) {
        System.out.println("The following host name cannot be resolved: "
            + args[i]);
        continue;
      }
    } // for

    /* Inizializzazione e attivazione dei multicast server */
    MulticastServer currentMCastServer;
    // ho controllato gli argomenti, posso istanziare i multicast server
    correctlyActivated = new boolean[validArgs];

    // inizializzo la struttura dati, ipotizzando che i server siano tutti
    // attivati correttamente
    for (int i = 0; i < validArgs; i++)
      correctlyActivated[i] = true;

    for (int i = 0; i < validArgs; i++) {
      try {
        currentMCastServer = new MulticastServer(validMCastAddresses[i],
            validFiles[i]);
        validPorts[i] = currentMCastServer.getPort();
        currentMCastServer.start();
      }
      catch (IOException e) {
        // Qui dovremmo mai passare, a meno di errori nel binding della
        // porta, o nello start del processo.
        // In tal caso dichiariamo il server che abbiamo attivato non
        // disponibile...
        correctlyActivated[i] = false;
        System.out
            .println("Discovery Server, exception during Multicast Server creation and activation, the following: "
                + e);
        e.printStackTrace();
      }
    }

    DatagramSocket socket = null;
    DatagramPacket packet = null;
    byte[] buf = new byte[256];

    try {
      // Nota: la porta si potrebbe anche passare come argomento
      // Provare a modificare il programma per realizzare cio'
      socket = new DatagramSocket(PORT);
      packet = new DatagramPacket(buf, buf.length);
      System.out.println("Creata la socket: " + socket);
    }
    catch (SocketException e) {
      System.out.println("Problemi nella creazione della socket: ");
      e.printStackTrace();
      System.exit(1);
    }

    try {
      String richiesta = null;
      ByteArrayInputStream biStream = null;
      DataInputStream diStream = null;
      StringTokenizer st = null;
      ByteArrayOutputStream boStream = null;
      DataOutputStream doStream = null;
      String risposta = null;
      byte[] data = null;

      while (true) {
        System.out.println("\nIn attesa di richieste...");

        // ricezione del datagramma
        try {
          packet.setData(buf);
          socket.receive(packet);
        }
        catch (IOException e) {
          System.err.println("Problemi nella ricezione del datagramma: "
              + e.getMessage());
          e.printStackTrace();
          continue;
          // il server continua a fornire il servizio ricominciando dall'inizio
          // del ciclo
        }

        try {
          biStream = new ByteArrayInputStream(packet.getData(), 0, packet
              .getLength());
          diStream = new DataInputStream(biStream);
          richiesta = diStream.readUTF();
          System.out.println("Richiesto server per nome file: " + richiesta);
        }
        catch (Exception e) {
          System.err.println("Problemi nella lettura della richiesta: ");
          e.printStackTrace();
          continue;
          // il server continua a fornire il servizio ricominciando dall'inizio
          // del ciclo
        }

        /* Generazione della risposta */
        risposta = null;
        for (int i = 0; i < validArgs; i++) {
          if (validFiles[i].toString().equals(richiesta)) {
            if (correctlyActivated[i]) {
              risposta = validMCastAddresses[i].getHostAddress() + " "
                  + validPorts[i];
              break;
            } else
              risposta = "File presente, ma multicast server non attivato";
          }
        }

        // problemi lato server, i.e. non abbiamo trovato il file
        if (risposta == null)
          risposta = "File non trovato";

        System.out.println("Risposta: " + risposta);

        // invio della risposta
        try {
          boStream = new ByteArrayOutputStream();
          doStream = new DataOutputStream(boStream);
          doStream.writeUTF(risposta);
          data = boStream.toByteArray();
          packet.setData(data, 0, data.length);
          socket.send(packet);
        }
        catch (IOException e) {
          System.err.println("Problemi nell'invio della risposta: "
              + e.getMessage());
          e.printStackTrace();
          continue;
          // il server continua a fornire il servizio ricominciando dall'inizio
          // del ciclo
        }

      } // while

    }
    // qui catturo le eccezioni non catturate all'interno del while
    // in seguito alle quali il server termina l'esecuzione
    catch (Exception e) {
      e.printStackTrace();
    }
    System.out.println("LineServer: termino...");
    socket.close();
    }

}