// MulticastClient.java

import java.io.*;
import java.net.*;
import java.util.StringTokenizer;

public class MulticastClient {

  public static void main(String[] args) {
    InetAddress addr = null;
    int port = -1;
    String fileName = null;
    try {
      if (args.length == 3) {
        addr = InetAddress.getByName(args[0]);
        port = Integer.parseInt(args[1]);
        fileName = args[2];
        System.out.println("Mi pongo in ascolto su:\nMCastAddr: " + args[0]
            + "\nMCastPort: " + port);
      } else {
        System.out
            .println("Usage: java MulticastClient ipDiscoveryServer portDiscoveryServer fileName");
        System.exit(1);
      }
    } //try
    catch (Exception e) {
      System.out.println("Problemi, i seguenti: ");
      e.printStackTrace();
      System.exit(1);
    }

    DatagramSocket datagramSocket = null;
    DatagramPacket datagramPacket = null;
    byte[] buf = new byte[256];

    // creazione della socket datagram, settaggio timeout di 30s
    // e creazione datagram packet
    try {
      datagramSocket = new DatagramSocket();
      datagramSocket.setSoTimeout(30000);
      datagramPacket = new DatagramPacket(buf, buf.length, addr, port);
      System.out.println("\nMulticastClient: avviato");
      System.out.println("Creata la socket: " + datagramSocket);

    }
    catch (SocketException e) {
      System.out.println("Problemi nella creazione della socket: ");
      e.printStackTrace();
      System.out.println("LineClient: interrompo...");
      System.exit(1);
    }

    InetAddress multicastAddr = null;
    int multicastPort = -1;
    ByteArrayOutputStream boStream = null;
    DataOutputStream doStream = null;
    String nomeFile = null;
    int numLinea = -1;
    String risposta = null;
    ByteArrayInputStream biStream = null;
    DataInputStream diStream = null;
    byte[] data = null;

    try {

      // riempimento e invio del pacchetto
      try {
        boStream = new ByteArrayOutputStream();
        doStream = new DataOutputStream(boStream);
        doStream.writeUTF(fileName);
        data = boStream.toByteArray();
        datagramPacket.setData(data);
        datagramSocket.send(datagramPacket);
        System.out.println("Richiesta inviata a " + addr + ", " + port);
      }
      catch (IOException e) {
        System.out.println("Problemi nell'invio della richiesta: ");
        e.printStackTrace();
      }

      try {
        // settaggio del buffer di ricezione
        datagramPacket.setData(buf);
        datagramSocket.receive(datagramPacket);
        // sospensiva solo per i millisecondi indicati, dopodich solleva una
        // SocketException
      }
      catch (IOException e) {
        System.out.println("Problemi nella ricezione del datagramma: ");
        e.printStackTrace();
      }

      // Estrazione risposta
      try {
        biStream = new ByteArrayInputStream(datagramPacket.getData(), 0, datagramPacket
            .getLength());
        diStream = new DataInputStream(biStream);
        risposta = diStream.readUTF();
        System.out.println("Risposta: " + risposta);
      }
      catch (IOException e) {
        System.out.println("Problemi nella lettura della risposta: ");
        e.printStackTrace();
      }

      /* Estrazione dei dati contenuti nella risposta */
      if (risposta.equals("File presente, ma multicast server non attivato")
          || risposta.equals("File non trovato")) {
        System.out.println("Errore, il seguente: ");
        System.out.println(risposta + "\nEsco...");
        datagramSocket.close();
        System.exit(4);
      } else {
        StringTokenizer st = new StringTokenizer(risposta);
        String ipGruppoMCast = st.nextToken();
        // DEBUG
        //System.out.println("IP multicast ricevuto: "+ipGruppoMCast);

        // non dovrebbe mai dare eccezione, perche' e' un indirizzo MCast corretto,
        // come verificato dal server
        multicastAddr = InetAddress.getByName(ipGruppoMCast);
        multicastPort = Integer.parseInt(st.nextToken());
        // liberiamo le risorse occupate
        datagramSocket.close();
      }
    }
    catch (Exception e) {
      System.out
          .println("Problemi nella ricezione dal Discovery Server: esco...");
      // si potrebbe gestire altrimenti l'eccezione, ad esempio tentando nuovamente
      e.printStackTrace();
      System.exit(5);
    }

    MulticastSocket multicastSocket = null;
    DatagramPacket multicastPacket = new DatagramPacket(buf, buf.length);

    try {
      multicastSocket = new MulticastSocket(multicastPort);
      multicastSocket.setSoTimeout(20000); // 20 secondi
      System.out.println("\nMulticastClient: avviato");
      System.out.println("Creata la socket: " + multicastSocket);
    }
    catch (IOException e) {
      System.out.println("Problemi nella creazione della socket: ");
      e.printStackTrace();
      System.exit(2);
    }

    // adesione al gruppo associato all'indirizzo multicast
    try {
      /* Per evitare problemi, con computer non connessi in rete,
       * bisogna impostare l'interfaccia di rete PRIMA di fare la
       * join al gruppo.
       *
       * VEDERE ANCHE LE FAQ nel sito!!!
       */
      // mi procuro l'inetaddress locale
      //InetAddress netInterface= InetAddress.getByName("localhost");
      // imposto l'interfaccia di rete
      //socket.setInterface(netInterface);
      multicastSocket.joinGroup(multicastAddr);
      System.out.println("Adesione al gruppo " + multicastAddr);
    }
    catch (IOException e) {
      System.out.println("Problemi nell'adesione al gruppo: ");
      e.printStackTrace();
      System.exit(3);
    }

    String linea = null;

    // ricezione di alcune linee
    for (int i = 0; i < 6; i++) {

      System.out.println("\nIn attesa di un datagramma... ");
      try {
        multicastPacket.setData(buf);
        multicastSocket.receive(multicastPacket);
      }
      catch (SocketTimeoutException te) {
        System.out.println("Non ho ricevuto niente per 20 secondi, chiudo!");
        System.exit(4);
      }
      catch (IOException e) {
        System.out.println("Problemi nella ricezione del datagramma: ");
        e.printStackTrace();
        // anche se ci sono problemi riprende il ciclo di ricezioni
        continue;
      }
      try {
        biStream = new ByteArrayInputStream(multicastPacket.getData(), 0, multicastPacket
            .getLength());
        diStream = new DataInputStream(biStream);
        linea = diStream.readUTF();
        System.out.println("Linea ricevuta: " + linea);
      }
      catch (IOException e) {
        System.out.println("Problemi nella lettura del datagramma: ");
        e.printStackTrace();
        continue;
        // anche se ci sono problemi riprende il ciclo di ricezioni
      }

    } //for

    // uscita dal gruppo e chiusura della socket
    System.out.println("\nUscita dal gruppo");
    try {
      multicastSocket.leaveGroup(multicastAddr);
    }
    catch (IOException e) {
      System.out.println("Problemi nell'uscita dal gruppo: ");
      e.printStackTrace();
    }

    System.out.println("MulticastClient: termino...");
    multicastSocket.close();
  }

}