// MulticastServer.java

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;

public class MulticastServer extends Thread {

  /*
   * Il file da dove leggo si deve trovare nella directory dove sono contenuti i
   * file .class
   * 
   */
  private static BufferedReader in = null;
  
  /*
   * Questa varibile di classe serve per ottenere una porta sulla quale
   * creare la socket
   * 
   */
  private static int listenPort = 4445;

  private static boolean moreLines = true;

  /* Indirizzo del gruppo multicast */ 
  private final InetAddress group;
  
  /* Socket */ 
  private MulticastSocket socket; 

  /* Porta */ 
  private int port;

  /* Pacchetto utilizzato per gli invii */ 
  private DatagramPacket packet;
  
  /* Array di byte utilizzato per gli invii */ 
  private byte[] data;
  

  /*
   * Il file da dove leggo si deve trovare nella directory dove sono contenuti i
   * file .class
   */
  private final File FILE;

  /**
   * Costruttore, con IP gruppo, porta e nome file da cui trasmettere.
   * 
   * Lancia eccezione se il file non esiste
   */
  public MulticastServer(InetAddress group, File FILE) throws IOException {
    super("MCServ: " + group);
    this.group = group;
    this.FILE = FILE;
    listenPort++;
    port = listenPort;
    data = new byte[256];
    packet = new DatagramPacket(data, data.length, group, port);

    // creazione della multicast socket e aggancio della multicast socket al
    // gruppo
    try {
      socket = new MulticastSocket(port);
      /*
       * Per evitare problemi, con computer non connessi in rete, bisogna
       * impostare l'interfaccia di rete PRIMA di fare la join al gruppo.
       * 
       * VEDERE ANCHE LE FAQ nel sito!!!
       */
      // mi procuro l'inetaddress locale
      //InetAddress netInterface= InetAddress.getByName("localhost");
      // imposto l'interfaccia di rete
      //socket.setInterface(netInterface);
      socket.joinGroup(group);
      System.out.println("Socket: " + socket);
    }
    catch (Exception e) {
      System.out.println("Problemi nella creazione della socket: ");
      e.printStackTrace();
      System.exit(1);
    }
  }

  public int getPort() {
    return socket.getLocalPort();
  }
  
  public void run() {
    BufferedReader in = null;
    boolean moreLines = true;

    long WAIT = 1000;


    System.out.println("MulticastServer: avviato\nMCastAddr: " + group
        + "\nport: " + socket.getLocalPort());

    int count = -1;
    ByteArrayOutputStream boStream = null;
    DataOutputStream doStream = null;

    while (true) {
      // azzero il contatore per il debug
      count = 0;
      moreLines = true;
      try {
        // costruisco un nuovo buffered reader
        in = new BufferedReader(new FileReader(FILE));
      }
      catch (Exception e) {
        System.out
            .println("MulticastServer, errore durante il reset del file reader, il seguente: "
                + e + "\nEsco.");
        System.exit(2);
        // provo di nuovo
        continue;
      }

      // eseguo il ciclo su tutte le linee del file
      while (moreLines) {

        count++;

        // estrazione della linea
        String linea = LineUtility.getNextLine(in);
        if (linea.equals("Nessuna linea disponibile")) {
          moreLines = false;
          // esco dal ciclo di lettura/trasmissione per
          // iniziarne un altro
          break;
        }

        //DEBUG
        // System.out.println("Estratta linea # " + count + " : " + linea);

        // costruzione del datagramma contenente la linea e invio della stessa
        try {
          boStream = new ByteArrayOutputStream();
          doStream = new DataOutputStream(boStream);
          doStream.writeUTF(linea);
          data = boStream.toByteArray();
          packet.setData(data);
          // invio della linea al gruppo
          socket.send(packet);
          //DEBUG
          // System.out.println("Linea inviata");
        }
        catch (Exception e) {
          System.out.println("Problemi nell'invio del datagramma: ");
          e.printStackTrace();
          // continuo a leggere la linea seguente
          continue;
        }

        // attesa tra un invio e l'altro...
        try {
          Thread.sleep((long) (Math.random() * WAIT));
        }
        catch (Exception e) {
          System.out
              .println("MulticastServer, errore durante la sleep, il seguente: "
                  + e);
          // continuo a leggere la linea seguente
          continue;
        } //catch
      } // while(moreLines)
    } // while(true)
  } //main

}